from pathlib import Path

import polars as pl
from sqlalchemy import create_engine, text

from ..utils import as_path

__all__ = ["create_db", "df_to_sql"]


# Definitions are fairly self-explanatory except for modified bytes. Modified bytes are calculated in two different
# ways:
#   1. modified_bytes_new: total storage used by the new version of the modified files
#   2. modified_bytes_net: net storage difference between old and new versions of modified files
CHURN_TBL_DEFINITION = """\
tld TEXT,
log_dt INTEGER,
prior_log_dt INTEGER,
created INTEGER,
created_bytes INTEGER,
deleted INTEGER,
deleted_bytes INTEGER,
modified INTEGER,
modified_bytes INTEGER,
modified_bytes_net INTEGER,
accessed INTEGER,
accessed_bytes INTEGER,
PRIMARY KEY (tld, log_dt)
"""

CHURN_TBL_COLS = [
    "created",
    "created_bytes",
    "deleted",
    "deleted_bytes",
    "modified",
    "modified_bytes",
    "modified_bytes_net",
    "accessed",
    "accessed_bytes",
]


def create_db(db_path: str | Path, table: str, definition: str):
    db_path = as_path(db_path)

    db_path.parent.mkdir(exist_ok=True, parents=True)

    engine = create_engine(f"sqlite:///{db_path}")
    with engine.connect() as conn:
        conn.execute(text(f"CREATE TABLE IF NOT EXISTS {table} ({definition})"))


def df_to_sql(df: pl.DataFrame, path: Path | str, table: str):
    path = as_path(path)
    uri = f"sqlite:///{path}"
    df.write_database(table, connection=uri, if_table_exists="append")


def create_churn_db(db_path: str | Path):
    table = "churn"
    definition = CHURN_TBL_DEFINITION
    create_db(db_path, table, definition)
