import json
import os
import random
import re
import shutil
import string
from pathlib import Path
from typing import List

import polars as pl

from ..utils import (
    as_bytes,
    as_path,
    calculate_age_distribution,
    calculate_size_distribution,
)


def collect_hive_df(
    parquet_path: str | Path,
    acq: str,
    tld: str | List[str] | None = None,
    hive_path: str | Path | None = None,
    no_clobber: bool = False,
) -> pl.DataFrame:
    if not isinstance(tld, list) and tld is not None:
        tld = [tld]

    print("Collecting dataframe", flush=True)

    df = (
        pl.scan_parquet(parquet_path, parallel="prefiltered", rechunk=True)
        .select(
            [
                "size",
                "kballoc",
                "access",
                "create",
                "modify",
                "uid",
                "gid",
                "path",
                "tld",
            ]
        )
        .with_columns(pl.lit(acq).alias("acq"))
    )

    if tld is not None:
        df = df.filter(pl.col("tld").is_in(tld))

    if no_clobber:
        existing_tlds = _get_existing_hive_cells(hive_path, acq)
        df = df.filter(pl.col("tld").is_in(existing_tlds).not_())

    df = df.collect(engine="streaming")

    print("Finished collecting queries", flush=True)

    return df


def _get_existing_hive_cells(hive_path: str | Path, acq: str):
    hive_path = as_path(hive_path)
    existing_pq = hive_path.rglob(f"*/acq={acq}/*.parquet")
    existing_tlds = list(
        set([p.parent.parent.name.removeprefix("tld=") for p in existing_pq])
    )
    return existing_tlds


def hivize(
    parquet_path: str | Path,
    hive_path: str | Path,
    tld: str | List[str] | None = None,
    partition_chunk_size_bytes: int | str = "200MiB",
    staging_path: str | Path | None = None,
    write_metadata: bool = True,
    no_clobber: bool = False,
    **kwargs,
) -> None:
    parquet_path = as_path(parquet_path).resolve()
    hive_path = as_path(hive_path).resolve()

    partition_chunk_size_bytes = int(as_bytes(partition_chunk_size_bytes))

    if staging_path is None:
        rand_str = "".join(random.choices(string.ascii_letters + string.digits, k=8))
        staging_path = Path(os.getenv("TMPDIR")).joinpath(
            os.getenv("USER"), f"hive-{rand_str}"
        )
        print(f"INFO: Using {staging_path} as temporary directory", flush=True)
    else:
        staging_path = as_path(staging_path)

    hive_path.mkdir(exist_ok=True, parents=True)
    staging_path.mkdir(exist_ok=True, parents=True)

    if tld is not None:
        if not isinstance(tld, list):
            tld = [tld]
        print(f"DEBUG: Hivizing {','.join(tld)}", flush=True)
    else:
        print("DEBUG: Hivizing all tlds", flush=True)

    acq = re.search(r"[\d]{4}-[\d]{2}-[\d]{2}", str(parquet_path)).group(0)
    print(f"DEBUG: Acquisition date is {acq}", flush=True)

    df = collect_hive_df(parquet_path, acq, tld, hive_path, no_clobber)

    if df.is_empty():
        print(
            "INFO: All passed tlds already have parquet files in their hive cell directories. Aborting",
            flush=True,
        )
        shutil.rmtree(staging_path)
        return

    print("Writing to hive")
    df.write_parquet(
        staging_path,
        partition_by=["tld", "acq"],
        partition_chunk_size_bytes=partition_chunk_size_bytes,
        statistics="full",
    )
    print("Finished writing hive dataset", flush=True)

    if write_metadata:
        tlds = df["tld"].unique().to_list()
        df = df.drop(
            [c for c in df.columns if c not in ["tld", "size", "modify", "access"]]
        ).lazy()
        for grp in tlds:
            tdf = df.filter(pl.col("tld").eq(grp))
            output_dir = staging_path.joinpath(f"tld={grp}", f"acq={acq}")
            write_dataset_metadata(tdf, output_dir, acq)

    shutil.copytree(staging_path, hive_path, dirs_exist_ok=True)
    shutil.rmtree(staging_path)


def write_dataset_metadata(
    df: pl.DataFrame | pl.LazyFrame, parquet_path: Path, acq: str, **kwargs
) -> dict:
    df = df.lazy()

    size_dist = (
        df.select("size")
        .with_columns(
            calculate_size_distribution(pl.col("size"), **kwargs).alias("grp")
        )
        .group_by("grp")
        .agg([pl.sum("size").alias("bytes"), pl.count("size").alias("file_count")])
        .sort("grp", descending=True)
        .collect(engine="streaming")
        .to_dicts()
    )

    access_dist = (
        df.select("access", "size")
        .with_columns(
            calculate_age_distribution(pl.col("access"), acq, **kwargs).alias("grp")
        )
        .group_by("grp")
        .agg([pl.sum("size").alias("bytes"), pl.count("size").alias("file_count")])
        .sort("grp", descending=True)
        .collect(engine="streaming")
        .to_dicts()
    )

    modify_dist = (
        df.select("modify", "size")
        .with_columns(
            calculate_age_distribution(pl.col("modify"), acq, **kwargs).alias("grp")
        )
        .group_by("grp")
        .agg([pl.sum("size").alias("bytes"), pl.count("size").alias("file_count")])
        .sort("grp", descending=True)
        .collect(engine="streaming")
        .to_dicts()
    )

    metadata = {
        "file_sizes": size_dist,
        "access_times": access_dist,
        "modify_times": modify_dist,
    }

    with open(parquet_path.joinpath("_metadata.json"), "w") as f:
        json.dump(metadata, f)
