from pathlib import Path
from typing import List, Literal

import numpy as np
import polars as pl
from typeguard import typechecked

from ..db.utils import CHURN_TBL_COLS
from ..utils import as_path, prep_age_distribution, prep_size_distribution

__all__ = ["aggregate_gpfs_dataset", "calculate_churn"]


def _check_dataset_path(dataset_path, file_glob) -> Path:
    dataset_path = as_path(dataset_path).resolve()

    if not dataset_path.exists():
        raise FileNotFoundError(f"{dataset_path} does not exist")

    if dataset_path.is_file():
        if file_glob is not None:
            print(f"INFO: {dataset_path} is a file, ignoring glob.")
            file_glob = None
        n_files = 1
    elif dataset_path.is_dir():
        n_files = len(list(dataset_path.glob(file_glob)))
        if n_files == 0:
            raise FileNotFoundError(
                f"No parquet files were found in {dataset_path.joinpath(file_glob)}"
            )
        dataset_path = dataset_path.joinpath(file_glob)

    print(f"INFO: {n_files} parquet file(s) found in {dataset_path}")

    return dataset_path


def _check_timedelta_values(vals, unit):
    if (vals is None) != (unit is None):
        print(
            "WARNING: time_breakpoints and time_unit were not both set. Skipping age aggregation"
        )
        vals = None
        unit = None
    return vals, unit


@typechecked
def aggregate_gpfs_dataset(
    df: pl.DataFrame | pl.LazyFrame,
    acq: str | np.datetime64,
    time_breakpoints: int | List[int] | None = [30, 60, 90, 180],
    time_unit: Literal["D", "W", "M", "Y"] | None = "D",
    time_val: Literal["access", "modify", "create"] | None = "access",
    size_breakpoints: int | str | List[int | str] | None = [
        "4 kiB",
        "4 MiB",
        "1 GiB",
        "10 GiB",
        "100 GiB",
        "1 TiB",
    ],
    **kwargs,
) -> pl.DataFrame:
    # Cast to LazyFrame to optimize aggregation. Is idempotent, will not affect a passed LazyFrame
    df = df.lazy()

    # Input checking
    time_breakpoints, time_unit = _check_timedelta_values(time_breakpoints, time_unit)

    grps = ["tld"]

    if time_breakpoints is not None:
        age_bins, age_labels = prep_age_distribution(acq, time_breakpoints, time_unit)

        df = df.with_columns(
            pl.col(time_val)
            .cut(breaks=age_bins, labels=age_labels)
            .cast(pl.String)
            .cast(pl.Enum(age_labels))
            .alias("dt_grp")
        )

        grps.append("dt_grp")

    if size_breakpoints is not None:
        size_bins, size_labels = prep_size_distribution(size_breakpoints)

        df = df.with_columns(
            pl.col("size")
            .cut(breaks=size_bins, labels=size_labels)
            .cast(pl.String)
            .cast(pl.Enum(size_labels))
            .alias("size_grp")
        )
        grps.append("size_grp")

    df_agg = (
        df.group_by(grps)
        .agg(
            [
                pl.col("size").sum().alias("bytes"),
                pl.col("size").count().alias("file_count"),
            ]
        )
        .sort(grps)
        .collect(engine="streaming")
    )

    return df_agg


@typechecked
def calculate_churn(
    hive_path: str | Path, tld: str, acq_dates: List[np.datetime64 | str], **kwargs
) -> pl.DataFrame:
    acq_dates = [
        np.datetime_as_string(d, "D") if isinstance(d, np.datetime64) else d
        for d in acq_dates
    ]
    acq_dates.sort()

    ## Input checking
    hive_path = as_path(hive_path)
    dataset_path = hive_path.joinpath(f"tld={tld}")
    acq_dirs = [d.name.removeprefix("acq=") for d in dataset_path.glob("acq=*")]

    # remove any datetimes for which the given tld does not have data.
    acq_dates = [d for d in acq_dates if d in acq_dirs]

    if len(acq_dates) <= 1:
        raise ValueError(
            f"Fewer than two given policy acquisition dates contained data for {tld} in {hive_path}."
        )

    df = (
        pl.scan_parquet(
            dataset_path.joinpath("**/*.parquet"),
            hive_partitioning=True,
            hive_schema=pl.Schema({"tld": pl.String, "acq": pl.String}),
            parallel="auto",
        )
        .select(["path", "acq", "access", "modify", "size"])
        .with_columns(pl.col("path").hash())
    )

    churn_l = []

    for i in range(1, len(acq_dates)):
        log_date = acq_dates[i]
        prior_date = acq_dates[i - 1]

        df1 = df.filter(pl.col("acq").eq(prior_date))
        df2 = df.filter(pl.col("acq").eq(log_date))

        churn = _calculate_churn(df1, df2)

        churn = churn.with_columns(
            pl.lit(log_date).alias("log_dt"),
            pl.lit(prior_date).alias("prior_log_dt"),
            pl.lit(tld).alias("tld"),
        )

        churn_l.append(churn)

    churn_l = pl.collect_all(churn_l, engine="streaming")
    churn_df = pl.concat(churn_l, how="vertical_relaxed").pivot(
        on="name", index=["tld", "log_dt", "prior_log_dt"], values="value"
    )

    return churn_df


def _calculate_churn(df1, df2) -> pl.DataFrame:
    empty_df = pl.LazyFrame(data={"name": CHURN_TBL_COLS, "value": 0})

    dfm = df1.join(df2, how="full", on="path", suffix="_updated", coalesce=True)

    conditions = [
        pl.col("access").is_null(),
        pl.col("access_updated").is_null(),
        pl.col("modify").ne(pl.col("modify_updated")),
        pl.col("access").ne(pl.col("access_updated")),
    ]

    choices = [
        pl.lit("created"),
        pl.lit("deleted"),
        pl.lit("modified"),
        pl.lit("accessed"),
    ]

    default = pl.lit("unchanged")

    dfm = dfm.with_columns(
        pl.when(conditions[0])
        .then(choices[0])
        .when(conditions[1])
        .then(choices[1])
        .when(conditions[2])
        .then(choices[2])
        .when(conditions[3])
        .then(choices[3])
        .otherwise(default)
        .cast(pl.Categorical)
        .alias("type")
    )

    dfm = dfm.filter(pl.col("type").ne("unchanged")).drop(
        ["modify", "access", "modify_updated", "access_updated"]
    )

    modified = dfm.filter(pl.col("type").eq("modified"))
    modified_bytes_net = modified.select(
        pl.lit("modified_bytes_net").alias("name"),
        (pl.sum("size_updated") - pl.sum("size")).alias("value"),
    )

    # Instead of writing logic to aggregate across initial size for deleted files and final size for all other
    # files, we can essentially condense size across both columns into a new column. Size of deleted files will
    # come from size while all other files will come from size_updated.
    dfm = (
        dfm.with_columns(pl.col("size_updated").fill_null(pl.col("size")))
        .drop("size")
        .rename({"size_updated": "size"})
    )

    agg_df = (
        dfm.group_by("type")
        .agg([pl.sum("size").alias("bytes"), pl.count("size").alias("files")])
        .unpivot(on=["bytes", "files"], index="type")
        .with_columns(
            pl.concat_str([pl.col("type"), pl.col("variable")], separator="_").alias(
                "name"
            )
        )
        .with_columns(pl.col("name").str.strip_suffix("_files"))
        .select(["name", "value"])
    )

    agg_df = (
        pl.concat([agg_df, modified_bytes_net], how="vertical_relaxed")
        .join(empty_df.select("name"), on="name", how="right")
        .fill_null(0)
        .sort("name")
    )

    return agg_df
