import pytest
from typing import Literal
from pathlib import Path
from rc_gpfs import utils

from polars.testing import assert_series_equal
import polars as pl
import numpy as np


### General Purpose Utils
@pytest.mark.parametrize("path", ["/data/rc/gpfs-policy", Path("/data/rc/gpfs-policy")])
def test_as_path_valid(path: Path | Literal["/data/rc/gpfs-policy"]):
    p_path = utils.as_path(path)
    assert isinstance(p_path, Path)


@pytest.mark.parametrize(
    "value,unit,to_unit,use_binary,expected",
    [
        (1, "G", "K", False, 1000000),
        (4, "K", "base", True, 4096),
        ("100", "base", "T", False, 1e-10),
    ],
)
def test_convert_si(value, unit, to_unit, use_binary, expected):
    assert utils.convert_si(value, unit, to_unit, use_binary) == expected


### Memory and File Size Utils
@pytest.mark.parametrize(
    "val,default,expected",
    [
        ("1 kiB", None, 1024),
        ("1 kiB", (1024**4), 1024),
        ("10 TiB", None, 10 * (1024**4)),
        (None, 1024, 1024),
        (None, None, None),
        ("1.5 MiB", None, 1572864),
    ],
)
def test_as_bytes(val, default, expected):
    assert utils.as_bytes(val, default) == expected


class TestSizeGrouping:
    input_sizes = pl.Series(
        name="size",
        values=[
            0,
            2048,
            4096,
            1024**2,  # 1 MiB
            1024**3,  # 1 GiB
            20 * 1024**3,  # 20 GiB
            1024**5,  # 1 PiB
        ],
        dtype=pl.Int128(),
    )

    expected_size_groups = [
        pl.Series(
            name="size",
            values=[
                "0 B-4 KiB",
                "0 B-4 KiB",
                "0 B-4 KiB",
                "4 KiB-4 MiB",
                "4 MiB-1 GiB",
                "10 GiB-100 GiB",
                ">1 TiB",
            ],
            dtype=pl.Enum(
                [
                    "0 B-4 KiB",
                    "4 KiB-4 MiB",
                    "4 MiB-1 GiB",
                    "1 GiB-10 GiB",
                    "10 GiB-100 GiB",
                    "100 GiB-1 TiB",
                    ">1 TiB",
                ]
            ),
        ),
        pl.Series(
            name="size",
            values=[
                "0 B-4 KiB",
                "0 B-4 KiB",
                "0 B-4 KiB",
                "4 KiB-4 MiB",
                "4 MiB-1 GiB",
                "10 GiB-100 GiB",
                ">1 TiB",
            ],
            dtype=pl.Enum(
                [
                    "0 B-4 KiB",
                    "4 KiB-4 MiB",
                    "4 MiB-1 GiB",
                    "1 GiB-10 GiB",
                    "10 GiB-100 GiB",
                    "100 GiB-1 TiB",
                    ">1 TiB",
                ]
            ),
        ),
        pl.Series(
            name="size",
            values=[
                "0 B-1 KiB",
                "1 KiB-4 KiB",
                "1 KiB-4 KiB",
                "4 KiB-10 GiB",
                "4 KiB-10 GiB",
                ">10 GiB",
                ">10 GiB",
            ],
            dtype=pl.Enum(["0 B-1 KiB", "1 KiB-4 KiB", "4 KiB-10 GiB", ">10 GiB"]),
        ),
        pl.Series(
            name="size",
            values=[
                "0 B-1 MiB",
                "0 B-1 MiB",
                "0 B-1 MiB",
                "0 B-1 MiB",
                ">1 MiB",
                ">1 MiB",
                ">1 MiB",
            ],
            dtype=pl.Enum(
                [
                    "0 B-1 MiB",
                    ">1 MiB",
                ]
            ),
        ),
    ]

    input_bins = [
        ["4 KiB", "4 MiB", "1 GiB", "10 GiB", "100 GiB", "1 TiB"],
        [4096, 4194304, 1073741824, 10737418240, 107374182400, 1099511627776],
        ["10 GiB", 1024, 4096, "1 KiB", 0],
        "1 MiB",
    ]

    expected_bins = [
        [4096, 4194304, 1073741824, 10737418240, 107374182400, 1099511627776],
        [4096, 4194304, 1073741824, 10737418240, 107374182400, 1099511627776],
        [1024, 4096, 10737418240],
        [1048576],
    ]

    expected_labels = [
        [
            "0 B-4 KiB",
            "4 KiB-4 MiB",
            "4 MiB-1 GiB",
            "1 GiB-10 GiB",
            "10 GiB-100 GiB",
            "100 GiB-1 TiB",
            ">1 TiB",
        ],
        [
            "0 B-4 KiB",
            "4 KiB-4 MiB",
            "4 MiB-1 GiB",
            "1 GiB-10 GiB",
            "10 GiB-100 GiB",
            "100 GiB-1 TiB",
            ">1 TiB",
        ],
        ["0 B-1 KiB", "1 KiB-4 KiB", "4 KiB-10 GiB", ">10 GiB"],
        ["0 B-1 MiB", ">1 MiB"],
    ]

    @pytest.mark.parametrize("bins,expected", list(zip(expected_bins, expected_labels)))
    def test_create_size_bin_labels(self, bins, expected):
        assert utils.create_size_bin_labels(bins) == expected

    @pytest.mark.parametrize(
        "bins,expected",
        list(zip(input_bins, list(zip(expected_bins, expected_labels)))),
    )
    def test_prep_size_distribution(self, bins, expected):
        assert utils.prep_size_distribution(size_bins=bins) == expected

    @pytest.mark.parametrize(
        "bins,expected", list(zip(input_bins, expected_size_groups)), ids=[1, 2, 3, 4]
    )
    def test_calculate_size_distribution(self, bins, expected):
        assert_series_equal(
            utils.calculate_size_distribution(self.input_sizes, size_bins=bins),
            expected,
        )


### DateTime and File Age Utils
@pytest.mark.parametrize(
    "date,expected",
    [
        ("2025-01-01", np.datetime64("2025-01-01T00:00:00.000000000")),
        (
            np.datetime64("2025-01-01T00:00:00.000000000"),
            np.datetime64("2025-01-01T00:00:00.000000000"),
        ),
        (1735689600000000000, 1735689600000000000),
        (None, None),
    ],
)
def test_as_datetime(date, expected):
    assert utils.as_datetime(date) == expected


def test_as_datetime_fails():
    with pytest.raises(ValueError):
        utils.as_datetime("not a date")


@pytest.mark.parametrize("val", [1, 3, 5])
@pytest.mark.parametrize("unit", ["D", "W"])
def test_as_timedelta(val, unit):
    assert utils.as_timedelta(val, unit) == np.timedelta64(val, unit)


class TestAgeGrouping:
    timestamps = pl.Series(
        name="access",
        values=[
            "2024-05-12",
            "2025-02-21",
            "2025-04-02",
            "2025-04-17",
            "2025-05-17",
        ],
    ).str.to_datetime(time_unit="ns")

    acq_date = "2025-06-01"

    delta_vals = [[-2, 30, 60, 90, 180], 365, [0, 4, 8, 12, 16], 52]

    delta_unit = ["D", "D", "W", "W"]

    expected_breakpoints = [
        [
            np.datetime64("2025-05-02T00:00:00.000000000"),
            np.datetime64("2025-04-02T00:00:00.000000000"),
            np.datetime64("2025-03-03T00:00:00.000000000"),
            np.datetime64("2024-12-03T00:00:00.000000000"),
        ],
        np.datetime64("2024-06-01T00:00:00.000000000"),
        [
            np.datetime64("2025-05-04T00:00:00.000000000"),
            np.datetime64("2025-04-06T00:00:00.000000000"),
            np.datetime64("2025-03-09T00:00:00.000000000"),
            np.datetime64("2025-02-09T00:00:00.000000000"),
        ],
        np.datetime64("2024-06-02T00:00:00.000000000"),
    ]

    expected_labels = [
        [">180D", "90D-180D", "60D-90D", "30D-60D", "<30D"],
        [
            ">365D",
            "<365D",
        ],
        [">16W", "12W-16W", "8W-12W", "4W-8W", "<4W"],
        [">52W", "<52W"],
    ]

    expected_age_groups_days = [
        pl.Series(
            name="access",
            values=[
                ">180D",
                "90D-180D",
                "60D-90D",
                "30D-60D",
                "<30D",
            ],
            dtype=pl.Enum([">180D", "90D-180D", "60D-90D", "30D-60D", "<30D"]),
        ),
        pl.Series(
            name="access",
            values=[">365D", "<365D", "<365D", "<365D", "<365D"],
            dtype=pl.Enum([">365D", "<365D"]),
        ),
        pl.Series(
            name="access",
            values=[">16W", "12W-16W", "8W-12W", "4W-8W", "<4W"],
            dtype=pl.Enum(
                [">16W", "12W-16W", "8W-12W", "4W-8W", "<4W"],
            ),
        ),
        pl.Series(
            name="access",
            values=[">52W", "<52W", "<52W", "<52W", "<52W"],
            dtype=pl.Enum([">52W", "<52W"]),
        ),
    ]

    @pytest.mark.parametrize(
        "delta_vals,delta_unit,expected",
        list(zip(delta_vals, delta_unit, expected_breakpoints)),
    )
    def test_create_timedelta_breakpoints_days(self, delta_vals, delta_unit, expected):
        assert (
            utils.create_timedelta_breakpoints(self.acq_date, delta_vals, delta_unit)
            == expected
        )

    @pytest.mark.parametrize(
        "delta_vals,delta_unit,expected",
        list(zip(delta_vals, delta_unit, expected_labels)),
    )
    def test_create_timedelta_labels(self, delta_vals, delta_unit, expected):
        assert utils.create_timedelta_labels(delta_vals, delta_unit) == expected

    @pytest.mark.parametrize(
        "age_breakpoints,delta_unit,expected",
        list(
            zip(
                delta_vals, delta_unit, list(zip(expected_breakpoints, expected_labels))
            )
        ),
    )
    def test_prep_age_distribution_days(self, age_breakpoints, delta_unit, expected):
        assert (
            utils.prep_age_distribution(self.acq_date, age_breakpoints, delta_unit)
            == expected
        )

    @pytest.mark.parametrize(
        "age_breakpoints,delta_unit,expected",
        list(zip(delta_vals, delta_unit, expected_age_groups_days)),
    )
    def test_calculate_age_distribution_days(
        self, age_breakpoints, delta_unit, expected
    ):
        assert_series_equal(
            utils.calculate_age_distribution(
                self.timestamps, self.acq_date, age_breakpoints, delta_unit
            ),
            expected,
        )
